/*
 * Copyright 2017-2019 original authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.micronaut.inject;

import io.micronaut.context.BeanResolutionContext;
import io.micronaut.context.exceptions.BeanInstantiationException;

import javax.validation.ConstraintViolation;
import javax.validation.Validator;
import javax.validation.ValidatorFactory;
import java.util.Optional;
import java.util.Set;

/**
 * A bean definition that is validated with javax.validation.
 *
 * @param <T> The bean definition type
 * @author Graeme Rocher
 * @since 1.0
 */
public interface ValidatedBeanDefinition<T> extends BeanDefinition<T> {

    /**
     * Validates the bean with the validator factory if present.
     *
     * @param resolutionContext The resolution context
     * @param instance          The instance
     * @return The instance
     */
    default T validate(BeanResolutionContext resolutionContext, T instance) {
        Optional<ValidatorFactory> validatorFactoryBean = resolutionContext.getContext().findBean(ValidatorFactory.class);
        if (validatorFactoryBean.isPresent()) {
            ValidatorFactory validatorFactory = validatorFactoryBean.get();
            Validator validator = validatorFactory.getValidator();
            Set<ConstraintViolation<T>> errors = validator.validate(instance);
            if (!errors.isEmpty()) {
                StringBuilder builder = new StringBuilder();
                builder.append("Validation failed for bean definition [");
                builder.append(instance.getClass().getName());
                builder.append("]\nList of constraint violations:[\n");
                for (ConstraintViolation<?> violation : errors) {
                    builder.append("\t").append(violation.getPropertyPath()).append(" - ").append(violation.getMessage()).append("\n");
                }
                builder.append("]");
                throw new BeanInstantiationException(resolutionContext, builder.toString());
            }
        }
        return instance;
    }
}
